// written by Zheng Gengbin
// deal with EXPRESSION, expression may be anywhere in any statement.
//

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
extern "C" void exit(...);
#endif

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#include "compilerApi.h"
#include "strategy.h"
#include "anls_att.h"


extern void prepare_actualdad_for_proc(SgStatement *s,SgExpression *var,int &arraycounter);
extern void change_procstmt(SgStatement *);

extern void sum_communication(enum Communication comm);
extern void change_oc_expression(SgStatement *,SgArrayRefExp *,int);

void inc_exp_level()
{
	int num=atoi(exp_level);
	num++;
	sprintf(exp_level, "%d", num);
//	exp_level[0]++;
}

void dec_exp_level()
{
	int num=atoi(exp_level);
	num--;
	sprintf(exp_level, "%d", num);
//	exp_level[0]--;
}

// analyse expression , extract element intrinsic function as stack 
// temp variable

#define INTRINSIC_FUNCTION  	1001
#define SCALAR_ARRAY_REF    	1002
#define SCALAR_POINTER_VARIABLE 1003

void collect_intrinsic(SgStatement *s,SgExpression *exp,int level)
{
	if (level ==0) s->deleteAllAttribute(INTRINSIC_FUNCTION);

	if (exp)
	switch (exp->variant() ) {
	case INTRIN_FUNC_CALL: {
//		if (((SgFunctionSymb*)exp->symbol())->is_elementary_function()==0)
//		{
//		s->addAttribute(INTRINSIC_FUNCTION,exp,sizeof(exp) );
			expression_reference *expr_ref = new expression_reference(exp);
			expr_ref->insert(s,INTRINSIC_FUNCTION);
//		}
		break;
	   }
	default: {
		SgExpression *lhs = exp->lhs();
		collect_intrinsic(s,lhs,level+1);
		SgExpression *rhs = exp->rhs();
		collect_intrinsic(s,rhs,level+1);
	   }
	}
}

void analyse_an_intrinsic(SgStatement *s,expression_reference *exp_ref,SgStatement *where)
{
	int j;
	SgExpression *exp = exp_ref->get_expression();
	SgExpression *same_expr = exp_ref->get_same_expression();

	// have same expression before
	if ( same_expr ) {
		exp->replace(same_expr->copy());
		return;
	}

	// already in NODE status.
	if (exp->haveAttribute(NODE_CODE_BIT)) return;

	SgFunctionCallExp *funccall = isSgFunctionCallExp(exp);
	SgType *type = exp->type();

	// we only deal with scalar function
	if (isSgArrayType(type) != NULL) return;

	  // if this intrinsic function is deal like functions.
	if (function_intrinsic(exp->symbol()) == 1 ) {
       		replace_intrinsic_function_name(exp,0);
		exp->setAttribute(NODE_CODE_BIT);
		return;
	}

  	  // if this is an elementary function, we ignore it , because we treat the 
	  // arg as an array in forall and assignment
	if (((SgFunctionSymb*)exp->symbol())->is_elementary_function()==1 )
	{
//		analyse_array_ref_in_exp(s,exp,s);
		return;
	}

if (!silentFlag)
printf("SCALAR_INTRINSIC_REF:%s\n",exp->symbol()->identifier());

       	// prepare dads for function
       	int arraycounter=1;	// must leave a space for the result
        // create dads
        for (j=0; j<funccall->numberOfArgs(); j++)
         	prepare_actualdad_for_proc(s,funccall->arg(j),arraycounter);

#ifdef MEM_ALLOC
    	SgSymbol * stack = s->getSymbolInCurrentScope(stackNameOfType(type),"mb", NULL); 
#else
    	SgSymbol * stack = s->getSymbolInCurrentScope(mixstring, stackNameOfType(type), NULL); 
#endif
	SgArrayRefExp * stackarray = new SgArrayRefExp(*stack);
	SgSymbol * basesym = s->getSymbolInCurrentScope(mixstring, "tbs", exp_level);
	SgExpression * baseexp= new SgVarRefExp(basesym);
    	stackarray->addSubscript(*baseexp);
	stackarray->setType(*type);

	SgAssignStmt *assign = new SgAssignStmt( *stackarray , *exp->copyPtr() );
       	s->insertStmtBefore(*assign,*s->controlParent());

       	intrinsic2sub(assign);     // change func_call to proc_call
       	replace_intrinsic_function_name(assign->expr(0),1);

	exp->replace(stackarray->copy());

	  // already in node program status
	exp->setAttribute(NODE_CODE_BIT);

       	SgSymbol *dadsec = s->getSymbolInCurrentScope(mixstring,"tdd",exp_level);
       	SgStatement *datastmt=make_new_array_data(s,dadsec,type);
       	assign->insertStmtBefore(*datastmt,*assign->controlParent());

	SgSymbol * mHandle;
#ifdef MEM_ALLOC
    	mHandle = s->getSymbolInCurrentScope(mixstring, "mh", exp_level);
	SgStatement *allocstmt=make_stack_alloc(s,type,basesym,dadsec,mHandle);
#else
	SgStatement *allocstmt=make_alloc(s,type,basesym,dadsec);
#endif
       	assign->insertStmtBefore(*allocstmt,*assign->controlParent());

	SgStatement *resetstmt=make_reset_array_base(s,dadsec,type,basesym,mHandle);
       	assign->insertStmtBefore(*resetstmt,*assign->controlParent());
		
#ifdef MEM_ALLOC
	basesym = s->getSymbolInCurrentScope(mixstring, "mh", exp_level);
#endif
       	SgStatement *freestmt=make_free(s,type,basesym);
       	where->insertStmtAfter(*freestmt,*assign->controlParent());

       	SgStatement *deletestmt=make_delete_array(s,dadsec);
       	assign->insertStmtAfter(*deletestmt,*assign->controlParent());

       	assign->expr(0)->lhs()->lhs()->addAttribute(DUMMY_ADDR,stackarray,sizeof(SgExpression *) );
       	assign->expr(0)->lhs()->lhs()->addAttribute(DUMMY_NAME,dadsec,sizeof(dadsec));
       	change_procstmt(assign);
       	assign->deleteStmt();

	// expression counter increase
	inc_exp_level();
}

void analyse_intrinsic_list(SgStatement *s,SgStatement *where)
{
	for (int i=0; i<s->numberOfAttributes(INTRINSIC_FUNCTION); i++) {
        	expression_reference *exp_ref = (expression_reference *)(s->attributeValue(i,INTRINSIC_FUNCTION));
		analyse_an_intrinsic(s,exp_ref,where);
	}
}

void analyse_intrinsic(SgStatement *s,SgExpression *expr,SgStatement *where)
{
	collect_intrinsic(s,expr,0);
	analyse_intrinsic_list(s,where);
}


/****************************************************************************
	analyse_array_ref_in_exp(SgStatement *s,SgExpression *expr,SgStatement *where,int counter)
	change array_ref in expression as a temporary stack
	array section creation is skipped !
***************************************************************************/

// collect all scalar array reference in expression
void collect_array_ref(SgStatement *s,SgExpression *exp,int level)
{
	if (level == 0)
	   s->deleteAllAttribute(SCALAR_ARRAY_REF);

	if (exp)
	switch (exp->variant() ) {
	case ARRAY_REF: {
		SgArrayRefExp *array = isSgArrayRefExp(exp);
		if (isSgArrayType(exp->type()) == NULL)
		{
			// recursively check every subscript
        		for (int i=0; i<array->numberOfSubscripts(); i++) 
        		{
          		   SgExpression *dimExp = array->subscript(i);
			   collect_array_ref(s,dimExp,level+1);
			}
			expression_reference *expr_ref = new expression_reference(exp);
			expr_ref->insert(s,SCALAR_ARRAY_REF);
		}
		break;
	   }
        case VAR_REF:  {
		break;
	   }
	default: {
		SgExpression *lhs = exp->lhs();
		collect_array_ref(s,lhs,level+1);
		SgExpression *rhs = exp->rhs();
		collect_array_ref(s,rhs,level+1);
	   }
	}
}

// caculate the offset of an array element.
SgExpression *get_scalar_subscripts(SgStatement *s,SgArrayRefExp *arrayexp)
{
	SgSymbol *arraysym = arrayexp->symbol();
        int storage = getArrayStorage(arraysym);

	// create subscripts
	SgExpression *sub = NULL;
	int numberOfSubscripts = arrayexp->numberOfSubscripts();
	int dim;
	for (dim=numberOfSubscripts-1; dim>=0 ; dim--)
	{
	   	int i;
	   	if (storage == COLUMN_MAJOR)
		   i = dim;
	   	else
		   i = numberOfSubscripts - dim -1;
		SgExpression *indexexp = arrayexp->subscript(i);
		SgExpression *one = new SgValueExp(1);
		indexexp = &( *indexexp - *one);
		// get the dim size of 
		SgExpression *dimsize = &( *get_ubound(arraysym,i) -*get_lbound(arraysym,i) + one->copy());
		dimsize = calculateExp(dimsize);
		if (sub==NULL) sub = indexexp;
		else sub = sub * *dimsize + *indexexp;
	}
	sub = calculateExp(sub);
	return sub;
}

// if an array scalar is collapsed , we caculate the global index expression.
void  analyse_collapse(SgStatement *s,SgArrayRefExp *arrayexp)
{
	SgSymbol *arraysym = arrayexp->symbol();
	SgExpression *sub = get_scalar_subscripts(s, arrayexp);
	int i;

//add by huang 98.3.11
        if(is_out_of_core(arraysym)==0){
                sub = new SgExpression(EXPR_LIST,sub,NULL,NULL);
                arrayexp->setSubscripts(*sub);
        }
        else{//case of out_of_core
		SgExpression * buf=new SgValueExp(atoi(exp_level));
		for(i=s->numberOfAttributes(OC_READ)-1;i>=0;i--){
		SgStatement * stmt=(SgStatement *)(s->attributeValue(i,OC_READ));
			if(stmt->variant()==PROC_STAT){
				SgCallStmt * stmt1=(SgCallStmt *)stmt; 
				SgArrayRefExp *argexp=(SgArrayRefExp *)stmt1->arg(1)->copyPtr();
				SgSymbol * argsym=argexp->symbol();
				if(argsym->identifier()==arraysym->identifier()){
//					if(*buf==* new SgValueExp(0)){ 
//						buf=&(*(argexp->subscript(0)->copyPtr())+*(new SgValueExp(1)));}
					if(*sub==*(stmt1->arg(2))){
						SgExpression * newsub=argexp->subscript(0)->copyPtr();	
               					newsub=new SgExpression(EXPR_LIST,newsub,NULL,NULL);
				                arrayexp->setSubscripts(*newsub);
						return;
					}	
				}
			}
		}
				
                SgSymbol *dad=s->getSymbolInCurrentScope(mixstring,"dad",
                        arraysym->identifier());
                SgSymbol *memsym=s->getSymbolInCurrentScope(arraysym->
                        identifier());
                SgArrayRefExp *memExp=new SgArrayRefExp(*memsym);
		buf=new SgExpression(EXPR_LIST,buf,NULL,NULL);
		memExp->setSubscripts(*buf);
                SgSymbol *stridesym=s->getSymbolInCurrentScope(mixstring,"oc_stride",arraysym->identifier());
                SgSymbol *numbersym=s->getSymbolInCurrentScope(mixstring,"oc_number",arraysym->identifier());
                SgStatement *ocreadStmt=make_oc_read(s,dad,memExp,sub,stridesym,numbersym);
                SgExpression *one=new SgValueExp(1);

                for(i=0;i<arrayexp->numberOfSubscripts();i++){
                        SgExpression *subscrpt=new SgValueExp(i+1);
                        subscrpt=new SgExpression(EXPR_LIST,subscrpt,NULL,NULL);                        SgArrayRefExp *numberexp=new SgArrayRefExp(numbersym->copy());
                        numberexp->setSubscripts(*subscrpt);
                        SgAssignStmt *assignStmt=new SgAssignStmt(*numberexp,
                                *one);
                        s->addAttribute(OC_READ,assignStmt,sizeof(assignStmt));
                }
		for(i=0;i<arrayexp->numberOfSubscripts();i++){
                        SgExpression *subscrpt=new SgValueExp(i+1);
                        subscrpt=new SgExpression(EXPR_LIST,subscrpt,NULL,NULL);                        SgArrayRefExp *stridexp=new SgArrayRefExp(stridesym->copy());
                        stridexp->setSubscripts(*subscrpt);
                        SgAssignStmt *assignStmt=new SgAssignStmt(*stridexp,
                                *one);
                        s->addAttribute(OC_READ,assignStmt,sizeof(assignStmt));
                }

                s->addAttribute(OC_READ,ocreadStmt,sizeof(ocreadStmt));

                buf=new SgExpression(EXPR_LIST,buf,NULL,NULL);
                arrayexp->setSubscripts(*buf);
		inc_exp_level();
        }
//add by huang finished!
}

void  analyse_collapse1(SgStatement *s,SgArrayRefExp *arrayexp)
{
	SgSymbol *arraysym = arrayexp->symbol();
	SgExpression *sub = get_scalar_subscripts(s, arrayexp);

//add by huang 98.3.11
	int i;
        if(is_out_of_core(arraysym)==0){
                sub = new SgExpression(EXPR_LIST,sub,NULL,NULL);
                arrayexp->setSubscripts(*sub);
        }
        else{//case of out_of_core
		SgExpression * buf=new SgValueExp(atoi(exp_level));
                SgSymbol *dad=s->getSymbolInCurrentScope(mixstring,"dad",
                        arraysym->identifier());
                SgSymbol *memsym=s->getSymbolInCurrentScope(arraysym->
                        identifier());
                SgArrayRefExp *memExp=new SgArrayRefExp(*memsym);
		buf=new SgExpression(EXPR_LIST,buf,NULL,NULL);
		memExp->setSubscripts(*buf);
                SgSymbol *stridesym=s->getSymbolInCurrentScope(mixstring,"oc_stride",arraysym->identifier());
                SgSymbol *numbersym=s->getSymbolInCurrentScope(mixstring,"oc_number",arraysym->identifier());
                SgStatement *ocwriteStmt=make_oc_write(s,dad,memExp,sub,stridesym,numbersym);
                s->addAttribute(OC_ASSIGN,ocwriteStmt, sizeof(ocwriteStmt));

                SgExpression *one=new SgValueExp(1);

                for(i=0;i<arrayexp->numberOfSubscripts();i++){
                        SgExpression *subscrpt=new SgValueExp(i+1);
                        subscrpt=new SgExpression(EXPR_LIST,subscrpt,NULL,NULL);                        SgArrayRefExp *numberexp=new SgArrayRefExp(numbersym->copy());
                        numberexp->setSubscripts(*subscrpt);
                        SgAssignStmt *assignStmt=new SgAssignStmt(*numberexp,
                                *one);
                        s->addAttribute(OC_ASSIGN, assignStmt, sizeof(assignStmt));
                }
		for(i=0;i<arrayexp->numberOfSubscripts();i++){
                        SgExpression *subscrpt=new SgValueExp(i+1);
                        subscrpt=new SgExpression(EXPR_LIST,subscrpt,NULL,NULL);                        SgArrayRefExp *stridexp=new SgArrayRefExp(stridesym->copy());
                        stridexp->setSubscripts(*subscrpt);
                        SgAssignStmt *assignStmt=new SgAssignStmt(*stridexp,
                                *one);
                        s->addAttribute(OC_ASSIGN, assignStmt, sizeof(assignStmt));
                }

                arrayexp->setSubscripts(*buf);
		inc_exp_level();
        }
//add by huang finished!
}

void  analyse_array_ref(SgStatement *s,SgExpression *exp,SgStatement *where,SgSymbol *sddsymb)
{
	SgArrayRefExp *arrayref = isSgArrayRefExp(exp);
	ASSERT(arrayref);
	SgType *type = exp->type();

#ifdef MEM_ALLOC
    	SgSymbol * stack = s->getSymbolInCurrentScope(stackNameOfType(type),"mb", NULL); 
#else
    	SgSymbol * stack = s->getSymbolInCurrentScope(mixstring, stackNameOfType(type), NULL); 
#endif
	SgArrayRefExp * stackarray = new SgArrayRefExp(*stack);
	SgSymbol * basesym = s->getSymbolInCurrentScope(mixstring, "tbs", exp_level);
	SgExpression * baseexp= new SgVarRefExp(basesym);
    	stackarray->addSubscript(*baseexp);
	stackarray->setType(*type);

        SgSymbol *dadsec = s->getSymbolInCurrentScope(mixstring,"tdd",exp_level);
        SgStatement *datastmt=make_new_array_data(s,dadsec,type);
        where->insertStmtBefore(*datastmt,*where->controlParent());

 	SgSymbol * mHandle;
#ifdef MEM_ALLOC
  	mHandle = s->getSymbolInCurrentScope(mixstring, "mh", exp_level );
	SgStatement *allocstmt=make_stack_alloc(s,type,basesym,dadsec,mHandle);
#else
	SgStatement *allocstmt=make_alloc(s,type,basesym,dadsec);
#endif
        where->insertStmtBefore(*allocstmt,*where->controlParent());

	SgStatement *resetstmt=make_reset_array_base(s,dadsec,type,basesym,mHandle);
        where->insertStmtBefore(*resetstmt,*where->controlParent());
	
//	SgSymbol *sddsymb=s->getSymbolInCurrentScope(mixstring, "sdd", exp_level );
	SgStatement *remapstmt = make_remap(s,dadsec,sddsymb);
       	where->insertStmtBefore(*remapstmt,*where->controlParent());

#ifdef MEM_ALLOC
	basesym = s->getSymbolInCurrentScope(mixstring, "mh", exp_level );
#endif
       	SgStatement *freestmt=make_free(s,type,basesym);
       	where->insertStmtAfter(*freestmt,*where->controlParent());

       	SgStatement *deletestmt=make_delete_array(s,dadsec);
       	where->insertStmtAfter(*deletestmt,*where->controlParent());

	exp->replace(*stackarray);

	inc_exp_level();
}

// if no need communication,return 1
// if symb is dummy, all-collapsed may be false, because the parent array
// may be dimension-reduced array, such as A(:,5) , when 5 is block distributed.
int expr_no_communication(SgStatement * s, SgSymbol *symb)
{
	if (is_resident(symb) == 1) return 1;

 	if (symb->numberOfAttributes(ALL_COLLAPSED) != 0)
		return 1;
	if (array_all_collapsed(symb) == 1) return 1;

/*	if (array_all_collapsed(symb) == 1)
	{
	    if (is_dummy(s, symb) == 0 ) 
	      	return 1;
	    else {
		 	if (symb->numberOfAttributes(ALL_COLLAPSED) != 0)
				return 1;
	         }
	}*/ //huang322
	return 0;
}

void  analyse_an_array_ref(SgStatement *s,expression_reference *exp_ref,SgStatement *where)
{
	SgExpression *expr = exp_ref->get_expression();
	SgExpression *same_expr = exp_ref->get_same_expression();

	// already in node program status
	if (expr->haveAttribute(NODE_CODE_BIT)) return;

	expr->setAttribute(NODE_CODE_BIT);

	// have same expression before
	if ( same_expr ) {
		expr->replace(same_expr->copy());
		return;
	}

	SgArrayRefExp *arrayexp = isSgArrayRefExp(expr);
	SgSymbol *arraysymb = arrayexp->symbol();
	ASSERT(arrayexp);
	SgType *type = expr->type();

if (!silentFlag)
printf("SCALAR_ARRAY_REF:%s -- ",arraysymb->identifier());

	// if array is collapsed and not dummy, no need to communicate
	// the actual parameter may be reduced dimension.
	if (expr_no_communication(s, arraysymb) == 1)
	{
if (!silentFlag)
printf("collapsed!.\n");
	if (debug_trans0) sum_communication(NONE);
	analyse_collapse(s,arrayexp);
	}
	else
	{
if (!silentFlag)
printf("remap!.\n");
	if (debug_trans0) sum_communication(REMAP);
	SgSymbol *dadsec=s->getSymbolInCurrentScope(mixstring, "sdd", exp_level);
	SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", arrayexp->symbol()->identifier());
	int dim =0;
	SgStatement *sectstmt=make_new_array_section(s, dadsec, dim , dad);
       	s->insertStmtBefore(*sectstmt,*s->controlParent());

	SgStatement *deletestmt=make_delete_array(s, dadsec);
       	where->insertStmtAfter(*deletestmt,*where->controlParent());

	define_array_from_arrayref(s,arrayexp, dadsec);

	analyse_array_ref(s,expr,where,dadsec);
	}
}

void  analyse_array_ref_list(SgStatement *s,SgStatement *where)
{
	int i;
	for (i=0; i<s->numberOfAttributes(SCALAR_ARRAY_REF); i++) 
	{
        	expression_reference *exp_ref = (expression_reference *)(s->attributeValue(i,SCALAR_ARRAY_REF));
		analyse_an_array_ref(s,exp_ref,where);
	}

//add by huang
	for (i=0; i<s->numberOfAttributes(OC_READ); i++) {
        	SgStatement *readst=(SgStatement *)s->attributeValue(i, OC_READ);
	
        	where->insertStmtBefore(*readst,*where->controlParent());
  	}
	s->deleteAllAttribute(OC_READ);
//add finished!	
}

// main function, analyse array refs in an complex expression
// s is the first statement and where is the last statement
void  analyse_array_ref_in_exp(SgStatement *s,SgExpression *exp,SgStatement *where)
{
	collect_array_ref(s,exp,0);
	analyse_array_ref_list(s,where);
}

void  analyse_array_ref_list_without_oc_read(SgStatement *s,SgStatement *where)
{
	int i;
	for (i=0; i<s->numberOfAttributes(SCALAR_ARRAY_REF); i++) 
	{
        	expression_reference *exp_ref = (expression_reference *)(s->attributeValue(i,SCALAR_ARRAY_REF));
		analyse_an_array_ref(s,exp_ref,where);
	}

}

// main function, analyse array refs in an complex expression
void  analyse_array_ref_in_exp_without_oc_read(SgStatement *s,SgExpression *exp,SgStatement *where)
{
	collect_array_ref(s,exp,0);
	analyse_array_ref_list_without_oc_read(s,where);
}

/****************************************************************************
	analyse_pointer_variable(SgStatement *s,SgExpression *expr,SgStatement *where,int counter)
	change pointer variable in expression to stack reference.
***************************************************************************/

// collect all scalar pointer variable in expression
void collect_scalar_pointer(SgStatement *s,SgExpression *exp,int level)
{
	if (level == 0)
	   s->deleteAllAttribute(SCALAR_POINTER_VARIABLE);

	if (exp)
	switch (exp->variant() ) {
	case ARRAY_REF: {
		break;
	   }
        case VAR_REF:  {
		SgSymbol *symb = exp->symbol();
		if (isSgArrayType(symb->type()) == NULL && is_pointer(symb))
             	{
		expression_reference *expr_ref = new expression_reference(exp);
		expr_ref->insert(s,SCALAR_POINTER_VARIABLE);
		}
		break;
	   }
	default: {
		SgExpression *lhs = exp->lhs();
		collect_array_ref(s,lhs,level+1);
		SgExpression *rhs = exp->rhs();
		collect_array_ref(s,rhs,level+1);
	   }
	}
}
	
void  analyse_an_pointer(SgStatement *s,expression_reference *exp_ref, SgStatement *where)
{
	SgExpression *expr = exp_ref->get_expression();
	SgExpression *same_expr = exp_ref->get_same_expression();

	// already in node program status
	if (expr->haveAttribute(NODE_CODE_BIT)) return;

	expr->setAttribute(NODE_CODE_BIT);

	// have same expression before
	if ( same_expr ) {
		expr->replace(same_expr->copy());
		return;
	}

	change_variable_by_stack(where, expr);
}

void  analyse_pointer_list(SgStatement *s,SgStatement *where)
{
	int i;
	for (i=0; i<s->numberOfAttributes(SCALAR_POINTER_VARIABLE); i++) 
	{
        	expression_reference *exp_ref = (expression_reference *)(s->attributeValue(i,SCALAR_POINTER_VARIABLE));
		analyse_an_pointer(s,exp_ref,where);
	}
}

// main function, analyse array refs in an complex expression
void  analyse_pointer_in_exp(SgStatement *s,SgExpression *exp,SgStatement *where)
{
	collect_scalar_pointer(s,exp,0);
	analyse_pointer_list(s,where);
}

/****************************************************************************
			analyse_expression()
 will do:
 1. analyse intrinsic functions.
 2. anaylse array reference.
****************************************************************************/

// first collect all expressions to process
void collect_expression(SgStatement *s, SgExpression *exp, int level)
{
	SgExpression *newexp = calculateExp(exp);
	if (newexp != exp) exp->replace(*newexp);

	collect_array_ref(s,exp,level++);
	collect_intrinsic(s,exp,level++);
	collect_scalar_pointer(s,exp,level++);
}

// second process each expressions 
// s is the first statement and where is the last
void analyse_expression_list(SgStatement *s, SgStatement *first, SgStatement *last)
{
	analyse_array_ref_list(s,last);
        analyse_intrinsic_list(s,last);
        analyse_pointer_list(s,s);
}

// collect and process expressions :exp
void analyse_expression(SgStatement *s, SgExpression *exp)
{
	analyse_array_ref_in_exp(s,exp,s);
	analyse_intrinsic(s,exp,s);
	analyse_pointer_in_exp(s,exp,s);
	  // calcaute expression if constant
	SgExpression *newexp = calculateExp(exp);
	if (newexp != exp) exp->replace(*newexp);
}

